#pragma TextEncoding="UTF-8"
#pragma rtGlobals=3
#pragma version=4.20
#pragma IgorVersion=7
#pragma ModuleName=tracer

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static constant kProjectID=342 // the project node on IgorExchange
static strconstant ksShortTitle="Tracer" // the project short title on IgorExchange

// https://www.wavemetrics.com/user/tony
// Please tell me when you find a bug, and let me know if you find this
// package useful or have suggestions for improvement.

// 4.20 A few minor changes, removed color-picker buttons because default
// color popup dialog has better selection tool with eye-dropper option. 
// Added headers for Updater. The rest of this code could use a clean-up!
// 4.10 Uses /any flag with loadimage to load more image types
// 4.09 Bug fix: data folder not returned to starting one after quit on load.
// 4.08 Added headers for Project Updater.
// See http://www.igorexchange.com/project/Updater.
// 4.07 bug fix: tracing failed with error when starting with cursor on a
// non-trace pixel
// 4.06 Oct 9 2017 Smaller panel size + bug fix
// 4.05 Sep 28 2017 New option to force search for bounds of trace over
// some minimum range. Use this if you're trying to trace unfilled
// symbols.
// 4.04 Sep 22 2017 bug fix. allowing gaps could sometimes result in
// non-integer point numbers, leading to out of range errors.
// 4.03 6/23/17 create copies of image for editing
// 4.02 8/25/16 bug fix for editing RGBA images with an alpha
// (transparency) channel
// 4.01 6/20/16 fixed null string bug for empty display list at startup
// 4.00 4/12/16 Greyscale (2d) images worked in previous versions because
// Igor didn't complain about out of range errors, so layer 1 and 2
// values were the same as layer 0. Fixed my bad code and introduced some
// new features. Created a "pixel wiper" to assist with touching up image
// to help with trace extraction. Image is assumed to be 3D RGB or 2D
// greyscale wave. Now works with 24 or 48 bit images (8 or 16 bits per
// pixel). A future update for Igor 7/8/9 will use ScaleToIndex, maybe
// ImageInterpolate Warp...

menu "Macros"
	"Tracer", /Q, tracer#Tracer_Initialise()
end

// sets up package folder and creates globals
// then creates plot and panel
static function Tracer_Initialise()
	
	NewDataFolder/O root:Packages
	NewDataFolder/O root:Packages:Tracer
	DFREF dfr=root:Packages:Tracer
	
	string /G dfr:NewTraceStr=""
	SVAR NewTraceStr = dfr:NewTraceStr
	NewTraceStr=UniqueName("trace", 1, 0)
	
	Make /O/N=11 dfr:sliderTicks=p/10
	Make /O/T/N=11 dfr:sliderLabels /wave=sliderLabels
	sliderLabels[0]="Exact"
	sliderLabels[10]="Fuzzy"
	
	Make /O/n=30  dfr:TracerGlobals /wave=g
	SetDimLabel 0, 0, null, g
	SetDimLabel 0, 1, XA, g
	SetDimLabel 0, 2, XB, g
	SetDimLabel 0, 3, YA, g
	SetDimLabel 0, 4, YB, g
	SetDimLabel 0, 5, gaps, g
	SetDimLabel 0, 6, JumpThreshold, g
	SetDimLabel 0, 7, R, g
	SetDimLabel 0, 8, G, g
	SetDimLabel 0, 9, B, g
	SetDimLabel 0, 10, deltaE, g
	SetDimLabel 0, 11, bgR, g
	SetDimLabel 0, 12, bgG, g
	SetDimLabel 0, 13, bgB, g
	SetDimLabel 0, 14, deltaEmax, g
	SetDimLabel 0, 15, mcsr, g
	SetDimLabel 0, 16, LogX, g
	SetDimLabel 0, 17, LogY, g
	SetDimLabel 0, 18, update, g
	SetDimLabel 0, 19, XYdata, g
	SetDimLabel 0, 20, editImg, g
	SetDimLabel 0, 21, wipe, g
	SetDimLabel 0, 22, UsageHints, g
	SetDimLabel 0, 23, forceRange, g
	SetDimLabel 0, 24, fuzzy, g
	SetDimLabel 0, 25, RGBthreshold, g
	SetDimLabel 0, 26, editCopies, g
	
	g[%null]=NaN
	g[%XA]=NaN
	g[%XB]=NaN
	g[%YA]=NaN
	g[%YB]=NaN
	g[%gaps]=0
	g[%JumpThreshold]=100
	g[%R]=0
	g[%G]=0
	g[%B]=0
	g[%deltaE]=10 // not used
	g[%bgR]=65535 // 16 bit RGB values
	g[%bgG]=65535
	g[%bgB]=65535
	g[%deltaEmax]=100 // not used
	g[%mcsr]=0
	g[%LogX]=0
	g[%LogY]=0
	g[%update]=50
	g[%XYdata]=0
	g[%editImg]=0
	g[%wipe]=0
	g[%UsageHints]=1
	g[%forceRange]=0 // seek for pixels over expanded range, even if a bounded pixel has been found
	g[%fuzzy]=0.2
	g[%RGBthreshold]=32*257 // ignore colour differences smaller than 12.5% in RGB space
	g[%editCopies]=1 // make a copy of image before editing
		// 0 = no copies, 1 = make copy, 2 (set by function) = check with user for continued editing
	
	DoWindow /K TracerGraph // start with a fresh window
	Display /K=1/N=TracerGraph /W=(200,10,550,350)  as "Tracer Image"
	Tracer_MakePanel()
end

static function Tracer_MakePanel()
	DFREF dfr=root:Packages:Tracer
	wave g=dfr:TracerGlobals
	
	DoWindow /K TracerPanel
	NewPanel /K=2/N=TracerPanel/W=(190,0,0,345)/HOST=TracerGraph/EXT=1 as "Tracer Controls"
	ModifyPanel /W=TracerGraph#TracerPanel, noEdit=1
	
	variable v_l=20, v_t=5, groupw=172, font=12, deltaY=25, i=0
	
	GroupBox groupImages, win=TracerPanel,pos={v_l,v_t},size={groupw,80},title="Image selection", fsize=font
	GroupBox groupImages, win=TracerPanel,pos+={-10,0}
	i+=1
	Button Load, win=TracerPanel,pos={25,v_t+i*deltaY},size={60,20},Proc=tracer#Tracer_Buttons,title="Load..."
	Button Load, win=TracerPanel,help={"Load an image to Packages:Tracer:"}, fsize=font
	Button clean, win=TracerPanel,pos={105,v_t+i*deltaY},size={60,20},title="Clear...", Proc=tracer#Tracer_Buttons
	Button clean, win=TracerPanel,help={"Kill all images in Packages:Tracer:"}, fsize=font
	i+=1
	PopupMenu popImage, win=TracerPanel, mode=1, Value=tracer#Tracer_ImgList(), pos={v_l,v_t+i*deltaY},size={190,20}
	PopupMenu popImage, win=TracerPanel, Proc=tracer#Tracer_PopMenuProc, title="Display: ", fsize=font
	PopupMenu popImage, win=TracerPanel, help={"Display an image from Packages:Tracer:" }
	i+=1.5
	
// set tabs
	TabControl tabScale, win=TracerPanel, tabLabel(0)="Scale",tabLabel(1)="Colour",tabLabel(2)="Extract"
	TabControl tabScale, win=TracerPanel, pos={v_l, v_t+i*deltaY},value=0,Proc=tracer#Tracer_Tab,size={groupw,160}
	TabControl tabScale, win=TracerPanel, pos+={-10,0}
	i+=1
	
//	tab0
	TitleBox titleX, win=TracerPanel,pos={69,v_t+i*deltaY},size={12,20},title="X",frame=0, fsize=font
	TitleBox titleY, win=TracerPanel,pos={126,v_t+i*deltaY},size={12,20},title="Y",frame=0, fsize=font
	i+=1
	TitleBox titleCsrA, win=TracerPanel,pos={18,i*deltaY},size={42,14},title="\\W642",frame=0, fsize=font+4
	SetVariable setvarXA, win=TracerPanel,pos={54,i*deltaY},size={43,16},title=" "
	SetVariable setvarXA, win=TracerPanel,limits={-Inf,Inf,0},value= g[%XA], fsize=font
	SetVariable setvarXA, win=TracerPanel,help={"Type in X value of cursor A"}
	SetVariable setvarYA, win=TracerPanel,pos={111,i*deltaY},size={43,16},title=" "
	SetVariable setvarYA, win=TracerPanel,limits={-Inf,Inf,0},value= g[%YA], fsize=font
	SetVariable setvarYA, win=TracerPanel,help={"Type in Y value of cursor A"}
	i+=1
	TitleBox titleCsrB, win=TracerPanel,pos={18,i*deltaY},size={42,14},title="\\W612",frame=0, fsize=font+4
	SetVariable setvarXB, win=TracerPanel,pos={54,i*deltaY},size={43,16},title=" "
	SetVariable setvarXB, win=TracerPanel,help={"Type in X value of cursor B"}
	SetVariable setvarXB, win=TracerPanel,limits={-Inf,Inf,0},value= g[%XB], fsize=font
	SetVariable setvarYB, win=TracerPanel,pos={111,i*deltaY},size={43,16},title=" "
	SetVariable setvarYB, win=TracerPanel,limits={-Inf,Inf,0},value= g[%YB], fsize=font
	SetVariable setvarYB, win=TracerPanel,help={"Type in Y value of cursor B"}
	i+=1
	CheckBox checkLogX, win=TracerPanel,pos={57,i*deltaY},size={36,14},Proc=tracer#Tracer_CheckProc,title="Log"
	CheckBox checkLogX, win=TracerPanel,value=g[%LogX], fsize=font
	CheckBox checkLogY, win=TracerPanel,pos={114,i*deltaY},size={36,14},Proc=tracer#Tracer_CheckProc,title="Log"
	CheckBox checkLogY, win=TracerPanel,value= g[%LogY], fsize=font
	i+=1
	Button setX, win=TracerPanel,pos={54,i*deltaY},size={43,20},Proc=tracer#Tracer_Buttons,title="Set X"
	Button setX, win=TracerPanel,help={"Set image X scaling"}, fsize=font
	Button setY, win=TracerPanel,pos={111,i*deltaY},size={43,20},Proc=tracer#Tracer_Buttons,title="Set Y"
	Button setY, win=TracerPanel,help={"Set image Y scaling"}, fsize=font
	
//	tab1
	i-=3.5
	PopupMenu popupColour, win=TracerPanel,pos={54,i*deltaY},size={50,21},Proc=tracer#Tracer_PopMenuProc, title="trace", bodywidth=50
	PopupMenu popupColour, win=TracerPanel,mode=1,popColor= (g[%R],g[%G],g[%B]),value= #"\"*COLORPOP*\"", fsize=font, disable=1
	PopupMenu popupColour, win=TracerPanel,help={"select \"Other...\", then the eye dropper to select color with mouse"}
	i+=1
	PopupMenu popupbg, win=TracerPanel,pos={54,i*deltaY},size={50,21},Proc=tracer#Tracer_PopMenuProc, title="bg", bodywidth=50
	PopupMenu popupbg, win=TracerPanel,mode=1,popColor= (g[%bgR],g[%bgG],g[%bgB]),value= #"\"*COLORPOP*\"", fsize=font, disable=1
	PopupMenu popupbg, win=TracerPanel,help={"select other, then the eye dropper to select color with mouse"}
	i+=1
	Slider slider0, win=TracerPanel,pos={18,i*deltaY},size={153,42},Proc=tracer#Tracer_SliderProc,fSize=font-2
	Slider slider0, win=TracerPanel,limits={0,1,0.05},vert= 0
	Slider slider0, win=TracerPanel,userTicks={dfr:sliderTicks,dfr:sliderLabels}
	Slider slider0, win=TracerPanel,value= g[%fuzzy], disable=1, help={"increase fuzziness to accept a greater range of pixel colors"}

//	tab2
	i-=2
	SetVariable setvarTraceName, win=TracerPanel,pos={18,i*deltaY},size={155,16},Proc=tracer#Tracer_SetVarProc,title="Create:"
	SetVariable setvarTraceName, win=TracerPanel,value= dfr:NewTraceStr, fsize=font, disable=1, help={"set name for new trace"}
	i+=1
	CheckBox checkXY, win=TracerPanel,pos={v_l,i*deltaY},size={58,14},Proc=tracer#Tracer_CheckProc,title="XY Data"
	CheckBox checkXY, win=TracerPanel,value= g[%XYdata], help={"create X and Y waves"}, fsize=font, disable=1
	CheckBox checkGaps, win=TracerPanel,pos={90,i*deltaY},size={71,14},Proc=tracer#Tracer_CheckProc,title="Allow Gaps", fsize=font
	CheckBox checkGaps, win=TracerPanel,value= g[%gaps], help={"allow tracing to continue when no target pixel is found"}, disable=1
	i+=1
	SetVariable setvarJump, win=TracerPanel,pos={25,i*deltaY},size={140,16},title="Jump threshold"
	SetVariable setvarJump, win=TracerPanel,limits={0,Inf,0},value= g[%JumpThreshold], fsize=font, disable=1
	SetVariable setvarJump, win=TracerPanel,help={"trace-colored pixels must be found within this range of\rprevious pixel to be considered part of continuous trace"}
	i+=1
	SetVariable setvarRange, win=TracerPanel,pos={25,i*deltaY},size={140,16},title="Minimum range"
	SetVariable setvarRange, win=TracerPanel,limits={0,Inf,0},value= g[%forceRange], fsize=font, disable=1
	SetVariable setvarRange, win=TracerPanel,help={"Force search to consider all pixels within range when\rsearching for bounds of trace (default = 0)"}
	i+=1
	Button Go, win=TracerPanel,pos={64,i*deltaY},size={50,20},Proc=tracer#Tracer_Buttons,title="Trace!", fsize=font, fColor=(57346,65535,49151), disable=1, help={"start tracing from cursor location"}

//	editing controls
	i+=1.3
	GroupBox groupEdit, win=TracerPanel,pos={10,i*deltaY},size={groupw,80},title="Edit", fsize=font
	i+=1
	Button editImg, win=TracerPanel,pos={50,i*deltaY},size={90,20},Proc=tracer#Tracer_Buttons,title="Edit image", fsize=font
	i+=1
	Button editTrace, win=TracerPanel,pos={50,i*deltaY},size={90,20},Proc=tracer#Tracer_Buttons,title="Edit trace", fsize=font
	
	// if we have already loaded an image, plot it
	ControlInfo /W=TracerGraph#TracerPanel popImage
	wave /Z w_img=dfr:$S_Value
	if (WaveExists(w_img))
		Tracer_PlotImage(w_img)
	endif
end

// get list of images in package datafolder
static function /S Tracer_ImgList()
	DFREF savDF=GetDataFolderDFR(), dfr=root:Packages:Tracer
	SetDataFolder dfr
	string ListOfImages=WaveList("*", ";", "DIMS:2")+WaveList("*", ";", "DIMS:3")
	SetDataFolder savDF
	if (strlen(ListOfImages)==0)
		ListOfImages="none;"
	endif
	return ListOfImages
end

static function Tracer_Buttons(s) : ButtonControl
	STRUCT WMButtonAction &s
	
	if (s.eventCode!=2)
		return 0
	endif
	
	DFREF dfr=root:Packages:Tracer
	wave g=dfr:TracerGlobals
	SVAR NewTraceStr = dfr:NewTraceStr

	string cmd, str
	
	strswitch(s.ctrlName)
		case "load":
			Tracer_LoadImage()
			return 1
		case "clean":
			Tracer_Cleanup()
			return 1
	endswitch
	
	if (strlen(ImageNameList("TracerGraph", ";" ))==0)
		DoAlert 0, "No image display!"
		return 0
	endif
	
	strswitch(s.ctrlName)
		case "setX":
			if (abs(g[%XA]-g[%XB])<1e-15)
				DoAlert 0, "Position cursors on X axis"
				return 0
			endif
			if (abs(g[%XA]==xcsr(A, "TracerGraph")) && (g[%XB]==xcsr(B, "TracerGraph")) )
				DoAlert 0, "First enter correct X values for cursor positions"
				return 0
			endif
			Tracer_SetImageScale(0)
			return 1
		case "setY":
			if (abs(g[%YA]-g[%YB])<1e-15)
				DoAlert 0, "Position cursors on Y axis"
				return 0
			endif
			if ( (g[%YA]==vcsr(A, "TracerGraph")) && (g[%YB]==vcsr(B, "TracerGraph")) )
				DoAlert 0, "First enter correct Y values for cursor positions"
				return 0
			endif
			Tracer_SetImageScale(1)
			return 1
		case "Go":
			Tracer_ExtractTrace()
			return 1
		case "editTrace":
			if (Tracer_EditTrace())
				Button editTrace, win=TracerGraph#TracerPanel, fcolor=(65535,0,0), title="Stop edit", Rename=StopEdit
				return 1
			endif
			return 0
		case "StopEdit":
			GraphNormal
			Button StopEdit, win=TracerGraph#TracerPanel, fColor=(0,0,0), title="Edit trace", Rename=editTrace
			return 1
		case "editImg":
			
			ControlInfo /W=TracerGraph#TracerPanel popImage
			variable MakeCopy=g[%editCopies]
			
			if (g[%editCopies])
				if (g[%editCopies]==2)
					DoAlert 2, "continue editing "+S_Value+"?\r\rNo will create a copy"
					if (V_flag==3)
						return 0
					elseif(V_flag==1)
						MakeCopy=0 // don't make a copy this time
					endif
				endif
			endif
			
			if(MakeCopy)
				wave /Z w_img=dfr:$S_Value
				if (WaveExists(w_img)==0)
					return 0
				endif
					
				DFREF savDF=GetDataFolderDFR()
				SetDataFolder dfr
				string newName=UniqueName(NameOfWave(w_img),1,0)
				SetDataFolder savDF
				Duplicate w_img dfr:$newName /wave=w
				
				// save current axis limits
				variable Lmin, Lmax, Bmin, Bmax
				GetAxis /W=TracerGraph/Q left
				Lmin=V_min;Lmax=V_max
				GetAxis /W=TracerGraph/Q bottom
				Bmin=V_min;Bmax=V_max
				
				PopupMenu popImage, win=TracerGraph#TracerPanel, popmatch=NameOfWave(w)
				Tracer_PlotImage(w)
				
				// restore axis limits
				SetAxis /W=TracerGraph left, Lmin, Lmax
				SetAxis /W=TracerGraph bottom, Bmin, Bmax
				
				g[%editCopies]=2
			endif
			
			if (g[%UsageHints])
				sprintf cmd, "Warning: you are about to edit %s\r\r", StringFromList(0, ImageNameList("TracerGraph", ";" ))
				cmd+="Hold ctrl/option and move mouse to add trace/bg colour. Expand view for detailed editing.\r\r"
				cmd+="Continue?"
				DoAlert 1, cmd
				if (v_flag==2)
					return 0
				endif
			endif
			Button editImg, win=TracerGraph#TracerPanel, fcolor=(65535,0,0), title="Stop edit", Rename=StopEditImg
			g[%editImg]=1
			SetWindow TracerGraph hook(EditImgHook)=tracer#Tracer_EditImgHook
			DoWindow /F TracerGraph
			return 0
		case "StopEditImg":
			Button StopEditImg, win=TracerGraph#TracerPanel, fColor=(0,0,0), title="Edit image", Rename=editImg
			g[%wipe]=0
			g[%editImg]=0
			SetWindow TracerGraph hook(EditImgHook)=$""
			return 1
	endswitch
end

static function Tracer_SliderProc(s) : SliderControl
	STRUCT WMSliderAction &s

	wave g=root:Packages:Tracer:TracerGlobals
	if(s.eventCode&1)	// bit 0, value set
		//g[%deltaE]=s.curval*g[%deltaEmax]
		g[%fuzzy]=s.curval
	endif
	return 0
end

static function Tracer_LoadImage()
	
	DFREF savDF=GetDataFolderDFR(), dfr=root:Packages:Tracer
	SetDataFolder dfr
	ImageLoad	/T=any
	if (V_Flag==0)
		SetDataFolder savDF
		return 0
	endif
	
	string imageStr=StringFromList(0, S_waveNames)	
	string newImageStr=ParseFilePath(3, imageStr, ":", 0, 0)
	newImageStr=CleanupName(newImageStr,0)
	if (CheckName(newImageStr, 1))
		newImageStr=UniqueName(newImageStr, 1, 0)
	endif
	Rename $imageStr $newImageStr
	SetDataFolder savDF
	wave /SDFR=dfr w_img=$newImageStr
	SetScale /P y, 0, -1, w_img
	Tracer_PlotImage(w_img)
		
	return 1
end

static function Tracer_PlotImage(w_img)
	wave w_img
	
	DFREF dfr=root:Packages:Tracer
	wave g=dfr:TracerGlobals
	
	// remove images and traces from plot
	Tracer_ClearPlot()
	
	AppendImage /W=TracerGraph w_img
	string str=NameOfWave(w_img)
	
	ModifyGraph width={Plan, abs(DimDelta(w_img, 1 )/ DimDelta(w_img, 0)) ,bottom,left}
	if(g[%LogX] || g[%LogY])
		Tracer_Replot()
	endif

	Cursor  /N=1/I/H=1/W=TracerGraph/C=(65535,0,0)/p A $str 0.1*DimSize(w_img, 0), 0.9*DimSize(w_img,1)
	Cursor  /N=1/I/H=1/W=TracerGraph/C=(0,65535,0)/p B $str 0.9*DimSize(w_img, 0), 0.9*DimSize(w_img,1)

	g[%XA]=xcsr(A, "TracerGraph")
	g[%XB]=xcsr(B, "TracerGraph")
	g[%YA]=vcsr(A, "TracerGraph")
	g[%YB]=vcsr(B, "TracerGraph")
	SetWindow TracerGraph hook(csrHook)=tracer#Tracer_csrHook
	
	// check this
	PopupMenu popImage, win=TracerGraph#TracerPanel, Value=tracer#Tracer_ImgList()
	PopupMenu popImage, win=TracerGraph#TracerPanel, popmatch=NameOfWave(w_img)
	
	if(g[%editCopies])
		g[%editCopies]=1 // reset when new image is displayed
	endif
	// if we clicked edit button this will be reset to 2 in Tracer_Buttons function
	
	return 1
end

static function Tracer_ClearPlot()
	string strImage="", strTrace=""
	do
		strImage=StringFromList (0, ImageNameList("TracerGraph", ";" ))
		RemoveImage /Z/W=TracerGraph $strImage
	while(strlen(strImage))
	do
		strTrace=StringFromList (0, TraceNameList("TracerGraph", ";" ,1))
		RemoveFromGraph /Z/W=TracerGraph $strTrace
	while(strlen(strTrace))
	return 1
end

static function Tracer_Cleanup()
	DoAlert 1, "Discard all tracer images from memory?"
	if (V_Flag==2)
		return 0
	endif
	
	Tracer_ClearPlot()
	DFREF savDF=GetDataFolderDFR(), dfr=root:Packages:Tracer
	SetDataFolder dfr
	
	string str, listofimages=WaveList("*", ";", "DIMS:2")+WaveList("*", ";", "DIMS:3")
	variable i
	for (i=0;i<ItemsInList(listofimages); i+=1)
		KillWaves /Z $StringFromList(i,listofimages)
	endfor
	SetDataFolder savDF
	ControlUpdate /W=TracerGraph#TracerPanel popImage
	return 1
end

static function Tracer_CheckProc(s) : CheckboxControl
	STRUCT WMCheckboxAction &s
	
	DFREF dfr=root:Packages:Tracer
	wave g=dfr:TracerGlobals
	
	if (s.eventCode!=2)
		return 0
	endif
					
	strswitch (s.ctrlName)
		case "CheckGaps":
			g[%gaps]=s.checked
			break
		case "CheckLogX":
			g[%LogX]=s.checked
			if (s.checked)
				g[%XYdata]=1
				CheckBox checkXY, win=TracerGraph#TracerPanel, value=1//, disable=2
			else
				g[%XYdata]=0
				CheckBox checkXY, win=TracerGraph#TracerPanel, value=0//, disable=0
			endif
			Tracer_Replot()
			break
		case "CheckLogY":
			g[%LogY]=s.checked
			Tracer_Replot()
			break
		case "CheckXY":
			g[%XYdata]=s.checked
			break
	endswitch
	return 0
end

// rebuild graph without changing image
// allows switch between log and normal axes
static function Tracer_Replot()
	DoWindow tracergraph
	if (V_Flag==0)
		return 0
	endif
	
	DFREF dfr=root:Packages:Tracer
	string strImage=StringFromList (0, ImageNameList("TracerGraph", ";" ))
	wave /Z w_image=ImageNameToWaveRef("TracerGraph",strImage)
	if (WaveExists(w_image)==0)
		return 0
	endif
	wave g=dfr:TracerGlobals
	
// save cursor positions and make sure they end up on the same pixels
	variable pcsrA, pcsrB, qcsrA, qcsrB
	if (stringmatch(StringByKey("TNAME", CsrInfo(A)), strImage) )// cursor is on image
		pcsrA=pcsr(A)
		qcsrA=qcsr(A)
	endif
	
	if (stringmatch(StringByKey("TNAME", CsrInfo(B)), strImage) )// cursor is on image
		pcsrB=pcsr(B)
		qcsrB=qcsr(B)
	endif
		
	RemoveImage /W=tracergraph  $strImage
	string ImageNote=note(W_Image)
	variable Xhigh, Xlow, Yhigh, Ylow
	
	if(g[%LogX])
		Make /O/n=(DimSize(w_image,0)+1) dfr:TracerLogX /wave=TracerLogX
		Xhigh=NumberByKey("Xhigh", ImageNote)
		if (numtype(Xhigh)==2)  //nan
			Xhigh=100
		endif
		Xlow=NumberByKey("Xlow", ImageNote)
		if (numtype(Xlow)==2)  //nan
			Xlow=1
		endif
		TracerLogX=alog(log(Xlow)+(p+0.5)/DimSize(TracerLogX,0)*(log(Xhigh)-log(Xlow)))
	endif
	if(g[%LogY])
		Make /O/n=(DimSize(w_image,1)+1) dfr:TracerLogY /wave=TracerLogY
		Yhigh=NumberByKey("Yhigh", ImageNote)
		if (numtype(Yhigh)==2)  //nan
			Yhigh=100
		endif
		Ylow=NumberByKey("Ylow", ImageNote)
		if (numtype(Ylow)==2)  //nan
			Ylow=1
		endif
		TracerLogY=alog(log(Ylow)+(DimSize(TracerLogY,0)-p-0.5)/DimSize(TracerLogY,0)*(log(Yhigh)-log(Ylow)))
	endif
	
	// need to specify this for Igor 7
	ModifyGraph  /W=TracerGraph width=0
		
	if(g[%LogX]&&g[%LogY])
		AppendImage /W=TracerGraph w_image vs {TracerLogX,TracerLogY}
	elseif(g[%LogX])
		AppendImage /W=TracerGraph w_image vs {TracerLogX,*}
	elseif(g[%LogY])
		AppendImage /W=TracerGraph w_image vs {*,TracerLogY}
	else
		AppendImage /W=TracerGraph w_image vs {*,*}
	endif

	ModifyGraph /W=TracerGraph log(left)=g[%LogY]
	ModifyGraph /W=TracerGraph log(bottom)=g[%LogX]
		
	if (g[%LogX] || g[%LogY])
		ModifyGraph  /W=TracerGraph width=0
	else
		ModifyGraph  /W=TracerGraph width={Plan, abs(DimDelta(w_image, 1 )/ DimDelta(w_image, 0)) ,bottom,left}
	endif
	
	if (numtype(pcsrA)==0 && numtype(qcsrA)==0)
		Cursor  /I/W=TracerGraph/p A $strImage pcsrA, qcsrA
	else
		Cursor  /I/W=TracerGraph/C=(0,65535,0)/p A $strImage 0.1*DimSize(W_Image, 0), 0.9*DimSize(W_Image,1)
	endif
	Cursor /M/C=(65535,0,0) /W=TracerGraph A
	
	if (numtype(pcsrB)==0 && numtype(qcsrB)==0)
		Cursor  /I/W=TracerGraph/p B $strImage pcsrB, qcsrB
	else
		Cursor  /I/W=TracerGraph/C=(0,65535,0)/p B $strImage 0.1*DimSize(W_Image, 0), 0.9*DimSize(W_Image,1)
	endif
		
	return 1
end

static function Tracer_SetVarProc(SV_Struct) : SetVariableControl
	STRUCT WMSetVariableAction &SV_Struct
	
	SVAR NewTraceStr = root:Packages:Tracer:NewTraceStr
	NewTraceStr=CleanupName(NewTraceStr, 0)
	return 1
end

static function Tracer_SetImageScale(dim)
	variable dim // x=0, y=1
	DFREF dfr=root:Packages:Tracer
	
	wave g=dfr:TracerGlobals
	
	variable A_val, B_val
	
	A_val=g[%YA]*(dim==1) + g[%XA]*(dim==0)
	B_val=g[%YB]*(dim==1) + g[%XB]*(dim==0)
	
	if (A_val==B_val)
		return 0
	endif
		
	wave pic=CsrWaveRef(A, "TracerGraph")
	variable A_P=(dim==0)*pcsr(A, "TracerGraph")+(dim==1)*qcsr(A, "TracerGraph")
	variable B_P=(dim==0)*pcsr(B, "TracerGraph")+(dim==1)*qcsr(B, "TracerGraph")
	variable V_delta=(A_Val-B_Val)/(A_P-B_P)
	
	string strNote=""
	variable Xhigh, Xlow, Yhigh, Ylow
	if ((dim==0)&&g[%LogX])
		if(A_val<=0 || B_val<=0)
			return 0
		endif
		strNote=note(pic)
		note /K pic
		
		wave TracerLogX=dfr:TracerLogX
		
		TracerLogX=alog(log(A_val)+(p-0.5-pcsr(a))*(log(B_val)-log(A_val))/(pcsr(B)-pcsr(A)))
		Xlow=alog(log(A_val)+(-1-pcsr(a))*(log(B_val)-log(A_val))/(pcsr(B)-pcsr(A)))
		Xhigh=alog(log(A_val)+(DimSize(TracerLogX,0)-1-pcsr(a))/(pcsr(B)-pcsr(A))*(log(B_val)-log(A_val)))
		strNote=ReplaceNumberByKey("XHigh", strNote, Xhigh)
		strNote=ReplaceNumberByKey("XLow", strNote, Xlow)
		note pic, strNote
		return 1
	endif
	if ((dim==1)&&g[%LogY])
		if(A_val<=0 || B_val<=0)
			return 0
		endif
		strNote=note(pic)
		note /K pic
		
		wave TracerLogY=dfr:TracerLogY
		
		TracerLogY=alog(log(A_val)+(p-0.5-qcsr(a))*(log(B_val)-log(A_val))/(qcsr(B)-qcsr(A)))
				
		// this part hurts my head.
		Ylow=alog(log(A_val)+(DimSize(TracerLogY,0)-1-qcsr(a))/(qcsr(B)-qcsr(A))*(log(B_val)-log(A_val)))
		Yhigh=alog(log(A_val)+(-1-qcsr(a))*(log(B_val)-log(A_val))/(qcsr(B)-qcsr(A)))
		
		strNote=ReplaceNumberByKey("YHigh", strNote, Yhigh)
		strNote=ReplaceNumberByKey("YLow", strNote, Ylow)
		note pic, strNote
		return 1
	endif
	
	wave w_image=CsrWaveRef(A, "TracerGraph")
	if (dim==0)
		SetScale /P x B_Val-V_delta*B_P, V_delta, w_image
	else
		SetScale /P y B_Val-V_delta*B_P, V_delta, w_image
	endif
	
	if (g[%LogX] || g[%LogY])
		ModifyGraph /W=TracerGraph width=0
	else
		ModifyGraph /W=TracerGraph width={Plan, abs(DimDelta(pic, 1 )/ DimDelta(pic, 0)) ,bottom,left}
	endif

	return 1
end

static function Tracer_EditTrace()
	string traceName=""
	if (!(ItemsInList( TraceNameList("TracerGraph",";",1))))
		GraphNormal // just in case
		DoAlert 0, "No traces on graph"
		return 0
	endif
	
	Prompt traceName, "Trace to edit:"Popup, TraceNameList("TracerGraph",";",1)
	DoPrompt "Select trace to edit", traceName
	
	if (V_flag)
		GraphNormal // just in case
		return 0
	endif
	
	GraphWaveEdit /W=TracerGraph /M $traceName
	return 1
end

static function Tracer_PopMenuProc(s) : PopupMenuControl
	STRUCT WMPopupAction &s
	if (s.eventCode!=2)
		return 0
	endif
	
	DFREF dfr=root:Packages:Tracer
	
	if (stringmatch(s.ctrlName, "popImage"))
		wave /Z w_img=dfr:$s.popStr
		if (WaveExists(w_img))
			Tracer_PlotImage(w_img)
		endif
		return 0
	endif
	
	// colour pickers
	
	ControlInfo /W=TracerGraph#TracerPanel $s.ctrlName
	wave /Z w_img=Tracer_ImgRef()
	wave g=dfr:TracerGlobals
	
	if (WaveExists(w_img) && (DimSize(w_img,2)==0))
		variable v_grey=(V_Red+V_Green+V_Blue)/3
		PopupMenu $s.ctrlName,win=TracerGraph#TracerPanel, popColor= (v_grey,v_grey,v_grey)
		ControlUpdate /W=TracerGraph#TracerPanel $s.ctrlName
		ControlInfo /W=TracerGraph#TracerPanel $s.ctrlName
	endif
	
	if (stringmatch(s.ctrlName, "popupColour"))
		g[%R]=V_Red
		g[%G]=V_Green
		g[%B]=V_Blue
	else // set bg colour
		g[%bgR]=V_Red
		g[%bgG]=V_Green
		g[%bgB]=V_Blue
	endif
	
	return 0
end

// this is where the magic happens
static function Tracer_ExtractTrace()
	DFREF dfr=root:Packages:Tracer
	wave g=dfr:TracerGlobals
	string ImageNameString=StringFromList (0, ImageNameList("TracerGraph", ";" ))
	wave /Z w_img=ImageNameToWaveRef("TracerGraph",ImageNameString)
	
	SVAR TraceNameString=dfr:NewTraceStr

	//ShowInfo /W=TracerGraph
	if ( strlen(CsrInfo(A, "TracerGraph"))==0 || strlen(CsrInfo(B, "TracerGraph"))==0 )
		DoAlert 0, "Set cursors at start and end of trace"
		return 0
	endif
	if (cmpstr(CsrWave(A, "TracerGraph"), CsrWave(B, "TracerGraph")))
		DoAlert 0, "Cursors must be on the same image"
		return 0
	endif
	if (pcsr(A, "TracerGraph")>pcsr(B, "TracerGraph"))
		variable tempP=pcsr(A, "TracerGraph"), tempQ=qcsr(A, "TracerGraph")
		Cursor /I /P /W=TracerGraph A $CsrWave(A, "TracerGraph") pcsr(B, "TracerGraph"), qcsr(B, "TracerGraph")
		Cursor /I /P /W=TracerGraph B $CsrWave(A, "TracerGraph") tempP, tempQ
	endif
	
	variable resolution=abs(pcsr(B, "TracerGraph")-pcsr(A, "TracerGraph"))+1
	
	TraceNameString=CleanupName(TraceNameString, 0 )
	if (CheckName(TraceNameString, 1) )
		DoAlert 1, TraceNameString+" already exists. Overwrite?"
		if (V_Flag==2)
			return 0
		endif
	endif
	
	Make /O/N=(resolution) $TraceNameString=NaN
	wave dataWave=$TraceNameString
	Print "created wave "+NameOfWave(dataWave)
	
	RemoveFromGraph /Z /W=TracerGraph $NameOfWave(dataWave) // just in case
	
	if (!g[%logX])
		SetScale /I x, xcsr(A, "TracerGraph"), xcsr(B, "TracerGraph"), dataWave
	endif
	
	if (g[%XYdata])
		Make /O/N=(resolution) $TraceNameString+"_X"=NaN
		wave dataXwave=$TraceNameString+"_X"
		SetScale /I x, xcsr(A, "TracerGraph"), xcsr(B, "TracerGraph"), dataXwave
		
		if (g[%logX]) // deal with X data for log X plots
			wave TracerLogX=dfr:TracerLogX
			dataXwave=TracerLogX[pcsr(A)+p+0.5]
		else
			dataXwave=x
		endif
	endif
		
	string info=ImageInfo("TracerGraph", ImageNameString, 0)
	
	if(g[%XYdata])
		AppendToGraph /W=TracerGraph dataWave vs dataXwave
	else
		AppendToGraph /W=TracerGraph dataWave
	endif
		
	ModifyGraph /W=TracerGraph mode($NameOfWave(dataWave))=0,lsize($NameOfWave(dataWave))=2, rgb($NameOfWave(dataWave))=(65280,0,0)
	// default tracing is red
	if (g[%R]>128*257 && (g[%G]+g[%B])<250*257) // data is red on image
		ModifyGraph /W=TracerGraph rgb($NameOfWave(dataWave))=(0,65280,0) // tracing will be green
	endif
	
	variable i=pcsr(A, "TracerGraph")
	variable j=0 // DataWave point number

	variable highP=qcsr(A, "TracerGraph"), lowP=qcsr(A, "TracerGraph")
	
	variable gap, jump=0, k, p_Y
	
	do
		gap=0
		if ((RGBgood(w_img, i, highP)==0)) // non-trace pixel
			do
				// expand to find a trace pixel
				lowP-=1
				highP+=1
							
				if (RGBgood(w_img, i, highP))
					lowP=highP
					break
				elseif(RGBgood(w_img, i, lowP))
					highP=lowP
					break
				endif
			while ( highP<(DimSize(w_img,1)-1) && lowP>1 )
		endif
		// check that we're on a trace pixel now
			
		if (highP>lowP) // failed to find one
			if (g[%gaps])
				gap=1
			else
				if (g[%XYdata])
					DoAlert 0, "Failed to follow trace at x=" +  num2str(dataXwave[j])+ "\rTry allowing gaps"
				else
					DoAlert 0, "Failed to follow trace at x=" +  num2str(pnt2x(dataWave, j))+ "\rTry allowing gaps"
				endif
				return 0
			endif
		else // found a trace pixel
			
			k=max(j,1)
			do
				k-=1 // find the last non-nan value in dataWave
			while((numtype(dataWave[k])==2)&&k)
			
			if (g[%LogY])
				wave TracerLogY=dfr:TracerLogY
				if (numtype(dataWave[k])==0)
					p_Y=BinarySearchInterp(TracerLogY, dataWave[k])
					// p_Y should be image column number

					if(abs(highP-p_Y)>g[%JumpThreshold])
						// more than JumpThreshold pixels away from previous value
						if (g[%gaps])
							gap=1 // treat it as a gap
						else
							if (g[%logX])
								wave TracerLogX=dfr:TracerLogX
								jump+=(jump==0)*TracerLogX[pcsr(A)+j+0.5]
							else
								jump+=(jump==0)*pnt2x(dataWave, j) // save first jump position
							endif
						endif
					endif
				endif
			elseif (abs(highP-(dataWave[k]-DimOffset(w_img, 1))/DimDelta(w_img, 1))>g[%JumpThreshold] )
				// more than JumpThreshold pixels away from previous value
				if (g[%gaps])
					gap=1 // treat it as a gap
				else
					if (g[%logX])
						wave TracerLogX=dfr:TracerLogX
						jump+=(jump==0)*TracerLogX[pcsr(A)+j+0.5]
					else
						jump+=(jump==0)*pnt2x(dataWave, j) // save first jump position
					endif
				endif
			endif
		endif
		if (gap)     // j is datawave point number, i is img x pixel
			
			k=max(j,1)
			do
				k-=1 // find the last non-nan value in dataWave
			while(numtype(dataWave[k])==2 && k)
			
			dataWave[j]=NaN
			
			if (numtype(datawave[k])) // must have failed to find first point
				highP=qcsr(A, "TracerGraph") // keep heading right until we hit the trace
			else
				if (g[%LogY])
					wave TracerLogY=dfr:TracerLogY
					highP=BinarySearch(TracerLogY, dataWave[k])
				else
					highP=round((dataWave[k]-DimOffset(w_img, 1))/DimDelta(w_img, 1))
				endif
			endif
			lowP=highP
		else		// not a gap
			if(g[%forcerange])
				highP+=floor(g[%forcerange]/2)
				lowP-=floor(g[%forcerange]/2)
			endif
			do // expand to find vertical limits of trace
				highP+=RGBgood(w_img,i,highP+1) // increment only if next pixel is good
				lowP-=RGBgood(w_img,i,lowP-1)
							
				lowP=max(lowP,0)
				highP=min(highP,DimSize(w_img,1)-1)
				
				if ( RGBgood(w_img,i,highP+1)==0 && RGBgood(w_img,i,lowP-1)==0 )
					// if we started on a trace pixel,
					// we've found the upper and lower bounds of trace
					break
				endif
			while (highP<(DimSize(w_img,1)-1) && lowP>0) // stop if we reach edge of image
			if(g[%forcerange]) // contract inward to find limits of trace
				do
					highP-=(RGBgood(w_img,i,highP)==0)
					lowP+=(RGBgood(w_img,i,lowP)==0)
				while( (RGBgood(w_img,i,highP)==0) || (RGBgood(w_img,i,lowP)==0) )
			endif

			// find the mid-point
			if (g[%LogY])
				wave TracerLogY=dfr:TracerLogY
				dataWave[j]=TracerLogY[(highP+LowP)/2+0.5]
			else
				dataWave[j]=DimOffset(w_img, 1)+DimDelta(w_img, 1)*(highP+LowP)/2
			endif
			
			highP=round((highP+LowP)/2)
			lowP=highP
		endif
		// allowing the doupdate can slow things down
		if((g[%update]) && mod(i, g[%update])==0) // sparse updating
			DoUpdate
		endif
		i+=1
		j+=1
	while(i<=pcsr(B, "TracerGraph"))
	
	// make sure we found something!
	variable badData=1
	for (j=0;j<DimSize(dataWave, 0); j+=1)
		if(numtype(dataWave[j])==0)
			badData=0
			break
		endif
	endfor
	
	if (jump!=0)
		Print "Failed to follow trace at x=" +  num2str(jump)
		DoAlert 1, "May have failed to follow trace at x=" +  num2str(jump)+ "\rTry allowing gaps or adjusting jump threshold\rDelete "+ NameOfWave(dataWave)+"?"
		if (V_flag==1)
			RemoveFromGraph /W=TracerGraph $NameOfWave(dataWave)
			KillWaves dataWave
			Print "deleted wave "+TraceNameString
		endif
	elseif ( badData || (abs(qcsr(B)- highP )>10) ) // no data, or more than 10 pixels away from target
		Print "Looks like trace failed"
		DoAlert  1, "Looks like trace failed.\rDelete "+ NameOfWave(dataWave)+"?"
		if (V_flag==1)
			RemoveFromGraph /W=TracerGraph $NameOfWave(dataWave)
			KillWaves dataWave
			Print "deleted wave "+TraceNameString
		endif
	else
		Print "Trace extracted successfully"
	endif
end

// determine out whether w_img[pixelX][pixelY] is close enough in color to be considered a trace pixel
static function RGBgood(w_img, pixelX, pixelY)
	wave w_img
	variable pixelX, pixelY	// image pixel
	
	DFREF dfr=root:Packages:Tracer
	wave g=dfr:TracerGlobals
	
//	// this sucked. I thought that a deltaE calculation would work better,
//	// but my made up method that's kinda linear in RGB space actually
//	// seems to work much better
//	w_RGBref={g[%R],g[%G],g[%B]}
//	// calculate Euclidian distance in L*ab colour space
//	return (Tracer_ColourDiff(w_RGB, w_RGBref)<=g[%deltaE])
	
//	// so try it this way
//	// calculate colours in CIE 1976 colour space
//	make /free/n=3 w_Lab, w_trLab, w_bgLab
//	Tracer_getRGB(w_Lab, w_img, pixelX, pixelY)
//	Tracer_RGB2LAB(w_Lab)
//	w_trLab={g[%R],g[%G],g[%B]}
//	Tracer_RGB2LAB(w_trLab)
//	w_bgLab={g[%bgR],g[%bgG],g[%bgB]}
//	Tracer_RGB2LAB(w_bgLab)
//	if (abs(w_Lab[0]-w_trLab[0])>max(g[%fuzzy]*10*(abs(w_trLab[0]-w_bgLab[0])), 5*g[%RGBthreshold]))
//		return 0
//	elseif (abs(w_Lab[1]-w_trLab[1])>max(g[%fuzzy]*(abs(w_trLab[1]-w_bgLab[1])), g[%RGBthreshold]))
//		return 0
//	elseif (abs(w_Lab[2]-w_trLab[2])>max(g[%fuzzy]*(abs(w_trLab[2]-w_bgLab[2])), g[%RGBthreshold]))
//		return 0
//	endif
//	return 1
		
	// do it the old way
	Make /free/n=3 w_RGB
	Tracer_getRGB(w_RGB, w_img, pixelX, pixelY)
	
	if (abs(w_RGB[0]-g[%R])>max(g[%fuzzy]*(abs(g[%R]-g[%bgR])), g[%RGBthreshold]))
		return 0
	elseif (abs(w_RGB[1]-g[%G])>max(g[%fuzzy]*(abs(g[%G]-g[%bgG])), g[%RGBthreshold]))
		return 0
	elseif (abs(w_RGB[2]-g[%B])>max(g[%fuzzy]*(abs(g[%B]-g[%bgB])), g[%RGBthreshold]))
		return 0
	endif
	return 1
end

static function Tracer_getRGB(w_RGB, w_img, pixelX, pixelY)
	wave w_RGB, w_img
	variable pixelX, pixelY
	
	if (DimSize(w_img,2))
		w_RGB=w_img[pixelX][pixelY][p]
	else
		w_RGB=w_img[pixelX][pixelY]
	endif
	
	// figure out colour depth
	// work with 24 or 48 bit images (8 or 16 bits per pixel)
	if (WaveType(w_img)&8)
		w_RGB*=257
	endif
	return 1
end

// write RGB values to a pixel, or a bunch of pixels,
// depending on how zoomed-in we are on the image
static function Tracer_writeRGB(w_RGB, w_img, pixelX, pixelY)
	wave w_RGB, w_img
	variable pixelX, pixelY
	
	DFREF dfr=root:Packages:Tracer
	wave g=dfr:TracerGlobals
	
	Duplicate /free w_RGB w
	// figure out colour depth
	// work with 24 or 48 bit images (8 or 16 bits per pixel)
	if (WaveType(w_img)&8)
		w/=257
	endif
	
	// figure out zoom level
	variable hpoints, vpoints
	GetAxis /W=TracerGraph /Q bottom
	if (g[%logX])
		wave TracerLogX=dfr:TracerLogX
		hpoints=abs(BinarySearch(TracerLogX, v_max)-BinarySearch(TracerLogX, v_min))
	else
		hpoints =abs(ScaleToIndex(w_img, v_max, 0)-ScaleToIndex(w_img, v_min, 0))		
	endif
	GetAxis /W=TracerGraph /Q left
	if (g[%logY])
		wave TracerLogY=dfr:TracerLogY
		vpoints=abs(BinarySearch(TracerLogY, v_max)-BinarySearch(TracerLogY, v_min))
	else
		vpoints =abs(ScaleToIndex(w_img, v_max, 1)-ScaleToIndex(w_img, v_min, 1))
	endif
	GetWindow tracergraph gsize
	variable zoom=hpoints/(v_right-v_left) // pixels per point
	zoom=min(zoom,vpoints/(v_bottom-v_top))
	zoom*=3
	
	variable pmin, pmax, qmin, qmax
	pmin=ceil(pixelX-zoom)
	pmax=floor(pixelX+zoom)
	qmin=ceil(pixelY-zoom)
	qmax=floor(pixelY+zoom)
	pmin=max(pmin, 0); qmin=max(qmin,0)
	pmax=min(pmax, (DimSize(w_img,0)-1))
	qmax=min(qmax, (DimSize(w_img,1)-1))
	variable i,j
	for (i=pmin;i<=pmax;i+=1)
		for (j=qmin;j<=qmax;j+=1)
			if (DimSize(w_img,2))
				w_img[i][j][0,2]=w[r]
			else
				w_img[i][j]=w[0]
			endif
		endfor
	endfor
	return 1
end

// hook function to deal with image editing
static function Tracer_EditImgHook(s)
	STRUCT WMWinHookStruct &s
	
	DFREF dfr=root:Packages:Tracer
	wave g=dfr:TracerGlobals
	
	if (g[%editImg]==0)
		g[%wipe]=0
		SetWindow TracerGraph hook(EditImgHook)=$""
		return 0
	endif
	
	wave w_img=Tracer_ImgRef()
	
	variable v_X, v_Y, p_X, p_Y
	
	variable v_key=GetKeyState(0)
	
	if ((g[%wipe]!=0) && s.eventCode==5)
		g[%wipe]=0  // mouseup cancels wipe
	endif
	// after mousedown g[%wipe] remembers status so that we can respond to mousemoved
	
	if (g[%editImg] && s.eventCode==3) // mousedown
		g[%wipe]=1
	endif
	
	// figure out if we want to add or erase trace colour from image
	variable v_insert=0 // -1 for erase, 1 for draw
	if (g[%editImg]&&v_key)
		// bit 4 is ctrl key on mac
		if ( (v_key&2^4) && (stringmatch(IgorInfo(2),"Macintosh")) )
			v_insert=1
			// set cursor to type 20
			s.doSetCursor=1
			s.cursorCode=20
		// bit 0 is ctrl on win
		elseif ( (v_key&2^0) && (stringmatch(IgorInfo(2),"Windows")) )
			v_insert=1
			s.doSetCursor=1
			s.cursorCode=20
		// bit 1 is option or alt
		elseif (v_key&2^1)
			v_insert=-1
			s.doSetCursor=1
			s.cursorCode=19
		endif
	else
		g[%wipe]=0 // stop wipe if no modifer keypress
		s.doSetCursor=0
	endif
	
	if ( g[%wipe] && (v_insert!=0) && (s.eventCode==3 || s.eventCode==4)  )
		// mousemoved or mousedown
		v_X=AxisValFromPixel("TracerGraph", "bottom", s.mouseLoc.h )
		v_Y=AxisValFromPixel("TracerGraph", "left", s.mouseLoc.v )
			
		// figure out which image pixel we're working on
		if(g[%logY])
			wave TracerLogY=dfr:TracerLogY
			p_Y=BinarySearch(TracerLogY, v_Y)
		else
			p_Y=ScaleToIndex(w_img, v_Y, 1)			
		endif

		if(g[%logX])
			wave TracerLogX=dfr:TracerLogX
			p_X=BinarySearch(TracerLogX, v_X)
		else			
			p_X=ScaleToIndex(w_img, v_X, 0)		
		endif

		if ( (0>p_X) || (p_X>(DimSize(w_img, 0)-1)) )
			return 0
		endif
		if ( (0>p_Y) || (p_Y>(DimSize(w_img, 1)-1)) )
			return 0
		endif
		
		Make /free/n=3 w_rgb
		if (v_insert==1) // set image pixel to trace colour
			w_rgb={g[%R],g[%G],g[%B]}
		else // set image pixel to bg colour
			w_rgb={g[%bgR],g[%bgG],g[%bgB]}
		endif
		Tracer_writeRGB(w_RGB, w_img, p_X, p_Y)
	endif
	
	return 0 // hook event labelled as not handled
	// allows nomal graph window interaction
end

// hook function to deal with cursors
static function Tracer_csrHook(s)
	STRUCT WMWinHookStruct &s
	
	if (s.eventCode!=7) // cursor moved
		return 0
	endif
	
	wave g=root:Packages:Tracer:TracerGlobals
	wave w_img=Tracer_ImgRef()
	
	Make /free /n=3 w_RGB
	
	strswitch(s.cursorName)
		case "A":
			if ( WaveRefsEqual(w_img,CsrWaveRef(A,"TracerGraph")) )
				g[%XA]=xcsr(a, "TracerGraph")
				g[%YA]=vcsr(a, "TracerGraph")
				
				// update cursor colour
				Tracer_GetRGB(w_RGB, w_img, pcsr(A,"TracerGraph"), qcsr(A,"TracerGraph") )
				if (w_RGB[0]>150*257 && (w_RGB[1]+w_RGB[2])<250*257) // reddish on image
					Cursor /M/C=(0,65535,0) /W=TracerGraph A // green cursor
				else
					Cursor /M/C=(65535,0,0) /W=TracerGraph A
				endif
			endif
			return 1
		case "B":
			if ( WaveRefsEqual(w_img,CsrWaveRef(B,"TracerGraph")) )
				g[%XB]=xcsr(B, "TracerGraph")
				g[%YB]=vcsr(B, "TracerGraph")
				
				// update cursor colour
				Tracer_GetRGB(w_RGB, w_img, pcsr(B,"TracerGraph"), qcsr(B,"TracerGraph") )
				if (w_RGB[1]>150*257 && w_RGB[0]+w_RGB[2]<250*257) // greenish on image
					Cursor /M/C=(0,0,65535) /W=TracerGraph B
				else
					Cursor /M/C=(0,65535,0) /W=TracerGraph B
				endif
			endif
			return 1
	endswitch
	return 0
end

// convert RGB to L*a*b* colour space
static function Tracer_RGB2LAB(w)
	wave w // 16 bit RGB values
	
	Make /n=3 /free w_XYZ, w_XYZr, w_f
	variable Ls, as, bs
	variable eps=216/24389
	variable k=24389/27
	
	// RGB to XYZ
	w=w/65535
	
	// assuming sRGB (D65)
	w = (w<=0.04045) ? (w/12.92) : ((w+0.055)/1.055)^2.4

// Bradford-adapted, D50 matrix
//	w_XYZ[0] = 0.436052025*w[0] + 0.385081593*w[1] + 0.143087414*w[2]
//	w_XYZ[1] = 0.222491598*w[0] + 0.71688606*w[1] + 0.060621486*w[2]
//	w_XYZ[2] = 0.013929122*w[0] + 0.097097002*w[1] + 0.71418547*w[2]
	
 	w_XYZ[0] = 0.4124564*w[0] + 0.3575761*w[1] + 0.1804375*w[2]
	w_XYZ[1] = 0.2126729*w[0] + 0.7151522*w[1] + 0.0721750*w[2]
	w_XYZ[2] = 0.0193339*w[0] + 0.1191920*w[1] + 0.9503041*w[2]
	
   // XYZ to Lab
   w_XYZr={0.964221,1,0.825211} // reference white D50
   w_XYZ=w_XYZ/w_XYZr
 
	w_f = (w_XYZ > eps) ? (w_XYZ^(1/3)) : (k*w_XYZ+16)/116
	Ls = 116*w_f[1] - 16
   as = 500*(w_f[0]-w_f[1])
   bs = 200*(w_f[1]-w_f[2])
   
//   w[0]=round(2.55*Ls + 0.5)
//   w[1]=round(as + 0.5)
//   w[2]=round(bs + 0.5)
  
   w={Ls,as,bs}
	return 1
end

// Computes the difference between two RGB colors
// by converting them to the L*a*b scale and
// comparing them using the CIE76 algorithm
// http://en.wikipedia.org/wiki/Color_difference#CIE76
static function Tracer_ColourDiff(w1, w2)
	wave w1, w2
	Duplicate /free w1 lab1
	Duplicate /free w2 lab2
	Tracer_RGB2LAB(lab1)
	Tracer_RGB2LAB(lab2)
	
	return sqrt((lab2[0]-lab1[0])^2+(lab2[1]-lab1[1])^2+(lab2[2]-lab1[2])^2)
end

static function /WAVE Tracer_ImgRef()
	string ImageNameString=StringFromList (0, ImageNameList("TracerGraph", ";" ))
	wave /Z w_img=ImageNameToWaveRef("TracerGraph",ImageNameString)
	return w_img
end

static function Tracer_Tab(tca) : TabControl
	STRUCT WMTabControlAction &tca

	switch( tca.eventCode )
		case 2: // mouse up
			variable tab = tca.tab
			
			DFREF dfr=root:Packages:Tracer
			wave g=dfr:TracerGlobals
			
			TitleBox titleX, win=$tca.win, disable=(tca.tab!=0)
			TitleBox titleY, win=$tca.win, disable=(tca.tab!=0)
			TitleBox titleCsrA, win=$tca.win, disable=(tca.tab!=0)
			SetVariable setvarXA, win=$tca.win, disable=(tca.tab!=0)
			SetVariable setvarYA, win=$tca.win, disable=(tca.tab!=0)
			TitleBox titleCsrB, win=$tca.win, disable=(tca.tab!=0)
			SetVariable setvarXB, win=$tca.win, disable=(tca.tab!=0)
			SetVariable setvarYB, win=$tca.win, disable=(tca.tab!=0)
			CheckBox checkLogX, win=$tca.win, disable=(tca.tab!=0)
			CheckBox checkLogY, win=$tca.win, disable=(tca.tab!=0)
			Button setX, win=$tca.win, disable=(tca.tab!=0)
			Button setY, win=$tca.win, disable=(tca.tab!=0)

			PopupMenu popupColour, win=$tca.win, disable=(tca.tab!=1)
			PopupMenu popupbg, win=$tca.win, disable=(tca.tab!=1)
			Slider slider0, win=$tca.win, disable=(tca.tab!=1)

			SetVariable setvarTraceName, win=$tca.win, disable=(tca.tab!=2)
			CheckBox checkXY, win=$tca.win, disable=(tca.tab!=2)
			
			if(g[%LogX] && tca.tab==2)
				CheckBox checkXY, win=$tca.win, disable=2
			endif
			
			CheckBox checkGaps, win=$tca.win, disable=(tca.tab!=2)
			SetVariable setvarJump, win=$tca.win, disable=(tca.tab!=2)
			SetVariable setvarRange, win=$tca.win, disable=(tca.tab!=2)
			Button Go, win=$tca.win, disable=(tca.tab!=2)

			break
		case -1: // control being killed
			break
	endswitch

	return 0
end